Package: Bioc.gff
Title: Read and write GFF and GTF files
Version: 1.3.0
Date: 2025-09-24
Authors@R: c(
    person("Michael", "Lawrence", role = "aut"),
    person("Hervé", "Pagès", role = "aut"),
    person("Marcel", "Ramos", role = "ctb"),
    person(
      "Bioconductor", "Package Maintainer", ,
      "maintainer@bioconductor.org", "cre"
    )
  )
Description: Parse GFF and GTF files using C++ classes. The package also
    provides utilities to read and write GFF3 files. The GFF (General Feature
    Format) format is a tab-delimited file format for describing genes and
    other features of DNA, RNA, and protein sequences. GFF files are often
    used to describe the features of genomes.
Depends: R (>= 4.5.0)
Imports: BiocBaseUtils, BiocGenerics, BiocIO, curl, GenomicRanges,
        IRanges, methods, Rsamtools, S4Vectors, Seqinfo, stats, utils,
        XVector
Suggests: BiocFileCache, BiocStyle, GenomicFeatures, GenomeInfoDbData,
        knitr, httr2, rmarkdown, rvest, tinytest, txdbmaker,
        TxDb.Hsapiens.UCSC.hg19.knownGene
VignetteBuilder: knitr
LinkingTo: S4Vectors, XVector, IRanges
License: Artistic-2.0
biocViews: Software, Infrastructure, DataImport
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
URL: https://github.com/Bioconductor/Bioc.gff
BugReports: https://github.com/Bioconductor/Bioc.gff/issues
Collate: 'Bioc.gff-package.R' 'GFF-coercion.R' 'readGFF.R'
        'GFFFile-class.R' 'index.R' 'metadataFromNCBI.R' 'utilities.R'
git_url: https://git.bioconductor.org/packages/Bioc.gff
git_branch: devel
git_last_commit: 8b9196c
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 21:30:03 UTC; biocbuild
Author: Michael Lawrence [aut],
  Hervé Pagès [aut],
  Marcel Ramos [ctb],
  Bioconductor Package Maintainer [cre]
Maintainer: Bioconductor Package Maintainer <maintainer@bioconductor.org>
