% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Bioc.gff-package.R
\docType{package}
\name{Bioc.gff-package}
\alias{Bioc.gff}
\alias{Bioc.gff-package}
\title{Bioc.gff: Read and write GFF and GTF files}
\description{
Parse GFF and GTF files using C++ classes. The package also provides utilities to read and write GFF3 files. The GFF (General Feature Format) format is a tab-delimited file format for describing genes and other features of DNA, RNA, and protein sequences. GFF files are often used to describe the features of genomes.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Bioconductor/Bioc.gff}
  \item Report bugs at \url{https://github.com/Bioconductor/Bioc.gff/issues}
}

}
\author{
\strong{Maintainer}: Bioconductor Package Maintainer \email{maintainer@bioconductor.org}

Authors:
\itemize{
  \item Michael Lawrence
  \item Hervé Pagès
}

Other contributors:
\itemize{
  \item Marcel Ramos [contributor]
}

}
\keyword{internal}
